# Copyright 2012 Leonidas Poulopoulos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from eos import eos
from pyparsing import *

class eosNed(eos):
    '''
    Subclass of eos to connect to the NE subsystem
    '''
    
    def __init__(self, server=None, port=None, password=None):
        '''
        Constructor for the class.      
        
        @type server: String
        @param server: The Hostname or IPv4 address of the Alcatel NMS.
        
        @type port: Integer
        @param port: The TCP port on the Alcatel NMS we will connect to. Default 20006
        
        @type password: String
        @param password: The password for this connection.
        
        @rtype: Handler
        @return: A connection handler of this class.
        '''        
        if port == None:
            port = 20006
        return eos.__init__(self, server, port, password)
    
    def list_ne_dir(self, parse=True):
        '''
        Function used to obtain a listing of of NE DATA from NMS. Depending on parse value the output is either raw data from NMS 
        or structured python data
        
        Usage:
            
            >>> from iooclient.eos_ned import *
            >>> ned = eosNed('1.2.3.4', 20006, 'password')
            >>> ned.list_ne_dir()
            [{'{'userLabel': 'elementName',
              'locationName': Athens,
              'notificationType': 'pmPointList',
              ...

        or:
            
            >>> from iooclient.eos_ned import *
            >>> with eosNed('1.2.3.4', 20006, 'password') as ned:
            >>>    ned.list_ne_dir()
            [{'{'userLabel': 'elementName',
              'locationName': Athens,
              'notificationType': 'pmPointList',
              ... 
                    
        @rtype: String or list of dictionaries
        @return: Network element data in raw text format or in structured format. False in case of errors.
        '''
        
        message = "LIST_CUR_NE_DIR_REQ[]"
        if self.send(message) == False:
            return False
        reply_header = self.get_part()
        if reply_header != "LIST_CUR_NE_DIR_CONF[]" :
            print "Error: Asked for %s but got unexpected output: %s\n" %(message, reply_header)
            return False
        reply = self.get()
        if reply == "DATA_END_NOTIF[]":
            return False
        if not parse:
            return reply
        ret = self.parseNeDir(reply)
        return ret
    
    def parseNeDir(self, data):
        ''' 
        Pyparsing nedir parser
        
        @type data: String
        @param data: Raw output from the NMS.
        
        @rtype: list of dicts
        @return: Transforms the NMS string into python list of dictionaries using the NMS delimiters
        '''
        escapechar = "\\"
        wordtext = CharsNotIn('\\*?^():"[]|=')
        escape = Suppress(escapechar) + (Word(printables, exact=1) | White(exact=1))
        wordvalues = Combine(OneOrMore(wordtext | escape | '=>'))
        rStart = Suppress('LIST_CUR_NE_DIR_CONF[]')
        rEnd = Suppress('DATA_END_NOTIF[]')
        response_body = Forward()
        response =  response_body + rEnd
        expr = Group(wordvalues + Suppress('=') + wordvalues)
        expr1 = Group( Suppress('NE_DIR_NOTIF') + Suppress('[')+Dict(delimitedList(expr, "|"))+Suppress(']'))
        expr2 = OneOrMore(expr1)
        response_body << expr2
        d = response.parseString(data)
        e = d.asList()
        return self.dictionarize(e)
    
    def dictionarize(self, list):
        '''
        Parses lists of lists of lists...etc into list of dictionaries of lists or dicts...etc
        
        @type list: List of lists of lists...
        @param list: Output from the pyparsing function
        
        @rtype: List
        @return: List of dicts of lists of dicts...etc
        '''
        l = []
        for item in list:
            l.append(dict(item))
        return l

