# Copyright 2012 Leonidas Poulopoulos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from eos import eos
import sys
from pyparsing import *

class eosPmd(eos):
    '''
    Subclass of eos to connect to the PM subsystem
    '''
    
    def __init__(self, server=None, port=None, password=None):
        '''
        
        Constructor for the class.

        @type server: String
        @param server: The Hostname or IPv4 address of the Alcatel NMS.
        
        @type port: Integer
        @param port: The TCP port on the Alcatel NMS we will connect to. Default 20007
        
        @type password: String
        @param password: The password for this connection.
        
        @rtype: Handler
        @return: A connection handler of this class.
        '''
        if port == None:
            port = 20007
        return eos.__init__(self, server, port, password)
    
    def list_pm_data(self, startPeriod=None, endPeriod=None, neNames=[], periodType=None, counters=[], parse=True):
        '''
        Function used to obtain a listing of of PM DATA from NMS. Depending on parse value the output is either raw data from NMS 
        or structured python data
        
        Usage:
        
            >>> from iooclient.eos_pmd import * 
            >>> pmd = eosPmd('1.2.3.4', 20007, 'password')
            >>> pmd.list_pm_data()
            [{'{'friendlyName': 'elementName',
              'neLocationName': Athens,
              'notificationType': 'pmPointList',
              ...
        
        or:
        
            >>> from iooclient.eos_pmd import * 
            >>> with eosPmd('1.2.3.4', 20007, 'password') as pmd:
            >>>    pmd.list_pm_data()
            [{'{'friendlyName': 'elementName',
              'neLocationName': Athens,
              'notificationType': 'pmPointList',
              ... 
        
        @type startPeriod: String
        @param startPeriod: From when to collect PM data. Format: YYYYMMDDQH where QH in [00,96] representing 15min periods in a day.
        
        @type endPeriod: String
        @param endPeriod: To when to collect PM data. Format: YYYYMMDDQH where QH in [00,96] representing 15min periods in a day.
        
        @type neNames: A list of complex strings
        @param neNames: A list of NetworkElement names with their respective pmPoints(optional) or {} if all pmPoints are assumed. To get different pmpoints from different network elements use the notation ['neName1{pmpoint1|pmpoint2}','neName2{pmpoint4|pmpoint5}','neName3{}']
                
        @type periodType: String
        @param periodType: The period of monitoring. Values: 15, 1h, 24, RM1, RM2, RM3, RM4, RM5, FL, immediate, peak, 30-day, IM, day, month. Default 15
        
        @type counters: A list of strings
        @param counters: The counters for the above NetworkElements to fetch. Assumes that the Network Elements have the same pmPoints. If omitted all counters are returned
        
        @type parse: Boolean
        @param parse: Default: True. If set to True, the output is a python list of dictionaries. Otherwise the output is the NMS response text.   
        
        @rtype: String or list of dictionaries
        @return: Performance monitoring data (counter values) in raw text format or in structured format. False in case of errors.
        '''
        extras = []
        extras.append('%s'%(periodType) if periodType else '15')
        extras.append('%s'%(startPeriod) if startPeriod else '')
        extras.append('%s'%(endPeriod) if endPeriod else '')
        extras.append('{%s}'%(('|'.join(counters)) if counters else ''))
        extras.append(''.join(["%s"%name for name in neNames]) if neNames else '')
        params = "|".join(extras) or ''
        message = "LIST_PM_DATA_REQ[%s]" %params
        if self.send(message) == False:
            return False
        reply_header = self.get_part()
        if reply_header != "LIST_PM_DATA_CONF[]" :
            print "Error: Asked for %s but got unexpected output: %s\n" %(message, reply_header)
            return False
        reply = self.get()
        if reply == "DATA_END_NOTIF[]":
            return False
        if parse:
            ret = self.parsePmData(reply)
        else:
            ret = reply
        return ret
    
    
    def list_pm_point(self, neNames=[], periodType=None, period=None, parse=True, listAll=False):
        '''
        Function used to obtain a listing of of PM points from NMS. Depending on parse value the output is either raw data from NMS 
        or structured python data 
        
        Usage:
            
            >>> from iooclient.eos_pmd import * 
            >>> pmd = eosPmd('1.2.3.4', 20007, 'password')
            >>> pmd.list_pm_point()
            [{'{'friendlyName': 'elementName',
              'neLocationName': Athens,
              'notificationType': 'pmPointList',
              ...
            
        or:
            
            >>> from iooclient.eos_pmd import * 
            >>> with eosPmd('1.2.3.4', 20007, 'password') as pmd:
            >>>    pmd.list_pm_point()
            [{'{'friendlyName': 'elementName',
              'neLocationName': Athens,
              'notificationType': 'pmPointList',
            ...
        
        @type neNames: A list of strings
        @param neNames: A list of NetworkElement names with pmPoints assigned. If omitted, all NEs are assumed
                
        @type periodType: String
        @param periodType: The period of monitoring. Values: 15, 1h, 24, RM1, RM2, RM3, RM4, RM5, FL, immediate, peak, 30-day, IM, day, month. Default 15
               
        @type period: String
        @param period: Period during which pm points were collected. Format: YYYYMMDDQH where QH in [00,96] representing 15min periods in a day.
        
        @type parse: Boolean
        @param parse: Default: True. If set to True, the output is a python list of dictionaries. Otherwise the output is the NMS response text.
        
        @type listAll: Boolean
        @param listAll: Default: False. If set to True, the output will be all points regardless of the default period set. Else, only the default period data will be fetched
        
        @rtype: String or list of dictionaries
        @return: Performance monitoring points (counter names) in raw text or in structured format. False in case of errors.
        '''
        extras = []
        extras.append('%s'%(periodType) if periodType else '15')
        extras.append('%s'%(period) if period else '')
        extras.append('{%s}'%(('|'.join(neNames)) if neNames else ''))
        params = "|".join(extras) or ''
        if listAll:
            params=''
        message = "LIST_PM_POINT_REQ[%s]" %params
        if self.send(message) == False:
            return False
        reply_header = self.get_part()
        if reply_header != "LIST_PM_POINT_CONF[]" :
            print "Error: Asked for %s but got unexpected output: %s\n" %(message, reply_header)
            return False
        reply = self.get()
        if reply == "DATA_END_NOTIF[]":
            return False
        if parse:
            ret = self.parsePmPoint(reply)
        else:
            ret = reply
        return ret
        
    def setGroupCounters(self, l, s, toks):
        ''' 
        Pyparsing action parser to include counters as key in pmPoint parsing
        '''
        return [["counters", toks[0][0].asList()]]
    
    def setGroupPmpoints(self, l, s, toks):
        ''' 
        Pyparsing action parser to include pmpoints as key in pmPoint parsing
        '''
        return [["pmpoints", toks[0][0].asList()]]
    
    def setGroupPeriods(self, l, s, toks):
        ''' 
        Pyparsing action parser to include periods as key in pmPoint parsing
        '''
        return [["periods", toks[0][0].asList()]]
    
    def parsePmPoint(self, data):
        ''' 
        Pyparsing pmpoint parser
        
        @type data: String
        @param data: Raw output from the NMS.
        
        @rtype: list of dicts
        @return: Transforms the NMS string into python list of dictionaries using the NMS delimiters
        '''
        escapechar = "\\"
        wordtext = CharsNotIn('\\*?^:"[]|={}')
        escape = Suppress(escapechar) + (Word(printables, exact=1) | White(exact=1))
        wordvalues = Combine(OneOrMore(wordtext | escape | '=>')).setParseAction(lambda t:t[0].replace('\xb0',''))
        rStart = Suppress('LIST_CUR_NE_DIR_CONF[]')
        rEnd = Suppress('DATA_END_NOTIF[]')
        response_body = Forward()
        response = response_body + rEnd
        pmCounters = Group(wordvalues + Suppress('=') + (wordvalues | (Suppress(Optional('{')) + Group(delimitedList(wordvalues, "|"))  + Suppress(Optional('}')))))
        pmPointBraces = Group(Suppress('{') + Dict(delimitedList(pmCounters, "|")) + Suppress('}'))
        pmPointsBraces = Group(Group(Suppress('{') + OneOrMore(pmPointBraces) + Suppress('}'))).setParseAction(self.setGroupPmpoints)
        pmPointInfo = Group(wordvalues + Suppress('=') + (wordvalues | Empty().setParseAction(replaceWith(None))))
        pmPointContent = Dict(delimitedList(pmPointInfo, "|")) + pmPointsBraces
        pmPointExpression = Group( Suppress('PM_POINT_NOTIF') + Suppress('[')+pmPointContent+Suppress(']'))
        allPoints = OneOrMore(pmPointExpression)
        response_body << allPoints
        d = response.parseString(data)
        e = d.asList()
        return parsePmpoints(e)
     
    def parsePmData(self, data):
        ''' 
        Pyparsing pmdata parser
        
        @type data: String
        @param data: Raw output from the NMS.
        
        @rtype: list of dicts
        @return: Transforms the NMS string into python list of dictionaries using the NMS delimiters
        '''
        escapechar = "\\"
        wordtext = CharsNotIn('\\*?^:"[]|={}')
        escape = Suppress(escapechar) + (Word(printables, exact=1) | White(exact=1))
        wordvalues = Combine(OneOrMore(wordtext | escape | '=>')).setParseAction(lambda t:t[0].replace('\xb0',''))
        rEnd = Suppress('DATA_END_NOTIF[]')
        response_body = Forward()
        response = response_body + rEnd
        equalityexpr = Group(wordvalues + Suppress('=') + (wordvalues | Empty().setParseAction(replaceWith(None))))
        dlist = Dict(delimitedList(equalityexpr, "|"))
        list = delimitedList(equalityexpr, "|")
        pmCounter = Suppress('{') + Group(Group(dlist)).setParseAction(self.setGroupCounters) + Suppress('}')
        pmCounterDates = Group(dlist + pmCounter)
        pmCounters = Group(Group(OneOrMore(pmCounterDates))).setParseAction(self.setGroupPeriods)
        pmCountersBraces = Suppress('{') + pmCounters + Suppress('}')
        pmPoint = Group(equalityexpr + pmCountersBraces)
        pmPoints = Group(Group(OneOrMore(pmPoint))).setParseAction(self.setGroupPmpoints)
        pmPointsBraces = Suppress('{') + pmPoints + Suppress('}')
        pmDataNotif = dlist + pmPointsBraces
        pmDataNotifExpr = Group( Suppress('PM_DATA_NOTIF') + Suppress('[')+pmDataNotif+Suppress(']'))
        pmDataAll = OneOrMore(pmDataNotifExpr)
        d = pmDataAll.parseString(data)
        e = d.asList()
        return parsePmpoints(e)

def isComplex(item):
    '''
    Simple expression to determine whether a variable is a complex (non-simple or string) list
    Complex list: [['a','b'], ['d','c']]
    Simple list: ['a', 'b', 'c']
    
    @type item: List or String
    @param item: The variable to be evaluated
    
    @rtype: Boolean
    @return: True if item is a complex list else False if item is simple list or string
    '''
    if isSimpleList(item) or isinstance(item, basestring):
        return False
    
    return True

def isSimpleList(item):
    '''
    Simple expression to determine whether a variable is a simple list
    Simple list: ['a', 'b', 'c']
    
    @type item: List or String
    @param item: The variable to be evaluated
    
    @rtype: Boolean
    @return: True if item is a simple list else False
    '''
    ret = False
    if isinstance(item, list):
        try:
            b = set(item)
            ret = True
        except:
            ret = False
    return ret

dictroot = {}
'''
Global dict variable that assists in transforming the pyparsing output 
to a python structured output
'''

def reDict(item):
    '''
    Multi-level dictionarize function. Transforms recursively lists and complex lists 
    into dictionaries
        
    @type item: List
    @param item: The variable to be evaluated
    
    @rtype: Dict
    @return: Dictionary of items
    '''
    if isSimpleList(item):
        return {item[0]: item[1]}
    dictroot = dict(item)
    for key,value in dictroot.items():
        if isinstance(value,list):
            if isComplex(value):
                dictroot[key] = [reDict(i) for i in value]
    return dictroot

def parsePmpoints(items):
    '''
    Parses lists of lists of lists...etc into list of dictionaries of lists or dicts...etc
    
    @type items: List of lists of lists...
    @param items: Output from the pyparsing function
    
    @rtype: List
    @return: List of dicts of lists of dicts...etc
    '''
    l = []
    for i in items:
        l.append(reDict(i))
    return l

